@@ -83,50 +83,13 @@ class Time:
                 self.shifts = np.random.permutation(self.shifts)
 
 
-class TimeOld:
-    def __init__(self, range_raw, level, variate_shift=False, permute_shifts=False):
-        self.range = cc_array.Range(range_raw)
-        self.level = level
-
-        self.permute_shifts = permute_shifts
-        self.variate_shift = variate_shift
-
-        if self.variate_shift:
-            self.S = abs(self.range.step)
-            if self.S == 1:
-                self.variate_shift = False
-                if self.permute_shifts:
-                    self.permute_shifts = False
-                self.shifts = np.asarray([0], dtype='i')
-
-            elif self.permute_shifts:
-                self.shifts = np.random.permutation(self.S)
-            else:
-                self.shifts = np.arange(self.S)
-        else:
-            self.S = 1
-            self.shifts = np.asarray([0], dtype='i')
-
-        self.s = 0
-        self.shift = self.shifts[self.s]
-
-    def next_shift(self):
-        if self.variate_shift:
-            if self.s < self.S - 1:
-                self.s += 1
-            else:
-                self.s = 0
-                if self.permute_shifts:
-                    self.shifts = np.random.permutation(self.S)
-            self.shift = self.shifts[self.s]
-
-
-class Loader:
+class LoaderOfBalancedAndClassDirectoryDataset:
 
     def __init__(
             self, directory_root, level_classes, n_levels_directories=None,
             conditions_directories=None, batch_size=None, n_batches=None,
-            shuffle=False, time=None, transforms=None, device=None):
+            shuffle=False, time=None, transforms=None, device=None,
+            return_combinations_eb=False, return_directories_eb=False):
 
         self.directory_root = directory_root
         self.level_classes = level_classes
@@ -147,6 +110,12 @@ class Loader:
         else:
             self.transforms = transforms
 
+        self.device = device
+
+        self.return_combinations_eb = return_combinations_eb
+
+        self.return_directories_eb = return_directories_eb
+
         self.conditions_directories_names = [None] * self.L  # type: list
         self.n_conditions_directories = [None] * self.L  # type: list
         self.conditions_directories = [None] * self.L  # type: list
@@ -172,6 +141,7 @@ class Loader:
             self.n_conditions_directories[self.time.level] = len(self.conditions_directories[self.time.level])
 
         self.K = self.n_classes = self.n_conditions_directories[self.level_classes]
+        self.V = self.n_variables = len(self.n_conditions_directories)
         self.n_samples = cc_maths.prod(self.n_conditions_directories)
         # self.n_samples = math.prod(self.n_conditions_directories)
 
@@ -189,7 +159,7 @@ class Loader:
         print('self.batch_size =', self.batch_size)
         print('self.n_samples =', self.n_samples)
         print('self.n_batches =', self.n_batches)
-        print('self.time.range_time =', self.time.range_time)
+        print('self.time =', self.time)
         print('self.n_conditions_directories =', self.n_conditions_directories)
         # print('self.conditions_directories =', self.conditions_directories)
 
@@ -198,7 +168,7 @@ class Loader:
         else:
             self.batches_indexes = np.split(np.arange(self.n_samples), self.n_batches, axis=0)
 
-        if self.time.random_shifts:
+        if (self.time is not None) and self.time.random_shifts:
             self.combinations_directories_no_shift = (
                 cc_combinations.conditions_to_combinations(self.conditions_directories))
             self.combinations_directories = None
@@ -211,7 +181,7 @@ class Loader:
             # self.combinations_directories[:, self.time.level] += self.time.shifts
             self.labels = torch.tensor(
                 self.combinations_directories_no_shift[slice(0, self.n_samples, 1), self.level_classes],
-                dtype=torch.int64, device=device)
+                dtype=torch.int64, device=self.device)
 
         else:
             self.combinations_directories_no_shift = None
@@ -220,7 +190,18 @@ class Loader:
 
             self.labels = torch.tensor(
                 self.combinations_directories[slice(0, self.n_samples, 1), self.level_classes],
-                dtype=torch.int64, device=device)
+                dtype=torch.int64, device=self.device)
+
+        # if self.return_combinations_eb:
+        #     self.combinations_eb = np.empty([self.batch_size, self.L], dtype='i')
+        self.labels_eb = None
+
+        self.combinations_eb = None
+
+        if self.return_directories_eb:
+            self.directories_eb = [None] * self.batch_size  # type: list
+        else:
+            self.directories_eb = None
 
         combination_directory_str_0 = [self.conditions_directories_names[l][0] for l in range(self.L)]
         directory_0 = os.path.join(self.directory_root, *combination_directory_str_0)
@@ -235,7 +216,7 @@ class Loader:
         shape_sample_0 = list(tensor_0.shape)
         self.shape_sample = shape_sample_0
         self.shape_batch = torch.Size(tuple([self.batch_size] + shape_sample_0))
-        self.tensor_batch = torch.empty(self.shape_batch, dtype=tensor_0.dtype, device=device)
+        self.tensor_batch = torch.empty(self.shape_batch, dtype=tensor_0.dtype, device=self.device)
 
         self.n_dims_samples = self.shape_sample.__len__()
         self.n_dims_batch = self.n_dims_samples + 1
@@ -255,169 +236,91 @@ class Loader:
         if self.shuffle:
             self.batches_indexes = np.split(np.random.permutation(self.n_samples), self.n_batches, axis=0)
 
-        for b in range(3):
-        # for b in range(self.n_batches):
-            labels_b = self.labels[self.batches_indexes[b]]
-            # batch_size_eb = batches_indexes_e.shape[0]
-            for i in range(self.batch_size):
-                combination_directory_ebi = self.combinations_directories[self.batches_indexes[b][i], :]
-                combination_directory_str_ebi = [
-                    self.conditions_directories_names[l][combination_directory_ebi[l]]
-                    for l in range(self.L)]
-                directory_ebi = os.path.join(self.directory_root, *combination_directory_str_ebi)
-                self.indexes_batch[0] = i
+        # self.n_batches = 3
+        if self.return_combinations_eb:
 
-                image_ebi = PIL.Image.open(directory_ebi)
-                self.tensor_batch[tuple(self.indexes_batch)] = self.transforms(image_ebi)
+            if self.return_directories_eb:
 
-            yield [self.tensor_batch, labels_b]
+                for b in range(self.n_batches):
+                    self.labels_eb = self.labels[self.batches_indexes[b]]
 
+                    self.combinations_eb = self.combinations_directories[self.batches_indexes[b], :]
+                    # batch_size_eb = batches_indexes_e.shape[0]
+                    for i in range(self.batch_size):
+                        combination_directory_ebi = self.combinations_eb[i, :]
+                        combination_directory_str_ebi = [
+                            self.conditions_directories_names[l][combination_directory_ebi[l]]
+                            for l in range(self.L)]
+                        directory_ebi = os.path.join(self.directory_root, *combination_directory_str_ebi)
 
-class LoaderOld:
+                        self.directories_eb[i] = directory_ebi
 
-    def __init__(
-            self, directory_root, level_classes, n_levels_directories=None,
-            conditions_directories=None, batch_size=None, n_batches=None,
-            shuffle=False, time=None, transforms=None, device=None):
+                        self.indexes_batch[0] = i
 
-        self.directory_root = directory_root
-        self.level_classes = level_classes
+                        image_ebi = PIL.Image.open(directory_ebi)
+                        self.tensor_batch[tuple(self.indexes_batch)] = self.transforms(image_ebi)
 
-        self.time = time
-
-        if n_levels_directories is None:
-            self.n_levels_directories = len(conditions_directories)
-        elif conditions_directories is None:
-            conditions_directories = [None] * self.n_levels_directories  # type: list
-
-        self.L = self.n_levels_directories
-
-        self.shuffle = shuffle
-
-        if transforms is None:
-            self.transforms = torchvision.transforms.Compose([torchvision.transforms.ToTensor()])
-        else:
-            self.transforms = transforms
-
-        self.conditions_directories_names = [None] * self.L  # type: list
-        self.n_conditions_directories = [None] * self.L  # type: list
-        self.conditions_directories = [None] * self.L  # type: list
-
-        directory_root_l = self.directory_root
-        for l in range(self.L):
-            self.conditions_directories_names[l] = os.listdir(directory_root_l)
-
-            if conditions_directories[l] is None:
-                self.n_conditions_directories[l] = len(self.conditions_directories_names[l])
-                self.conditions_directories[l] = np.arange(self.n_conditions_directories[l])
+                    yield [self.tensor_batch, self.labels_eb, self.combinations_eb, self.directories_eb]
             else:
-                self.conditions_directories[l] = conditions_directories[l]
-                self.n_conditions_directories[l] = len(self.conditions_directories[l])
 
-            directory_root_l = os.path.join(directory_root_l, self.conditions_directories_names[l][0])
+                for b in range(self.n_batches):
+                    self.labels_eb = self.labels[self.batches_indexes[b]]
 
-        if self.time is not None:
-            # if self.time.T is None:
-            #     self.time.T = self.n_conditions_directories[self.time.level]
-            if self.time.range.stop is None:
-                self.time.range.stop = self.n_conditions_directories[self.time.level]
-            if self.time.level is None:
-                self.time.range.level = self.L - 1
+                    self.combinations_eb = self.combinations_directories[self.batches_indexes[b], :]
+                    # batch_size_eb = batches_indexes_e.shape[0]
+                    for i in range(self.batch_size):
+                        combination_directory_ebi = self.combinations_eb[i, :]
+                        combination_directory_str_ebi = [
+                            self.conditions_directories_names[l][combination_directory_ebi[l]]
+                            for l in range(self.L)]
 
-            self.conditions_directories[self.time.level] = (
-                self.conditions_directories[self.time.level][self.time.range.to_slice()])
-            self.n_conditions_directories[self.time.level] = len(self.conditions_directories[self.time.level])
+                        directory_ebi = os.path.join(self.directory_root, *combination_directory_str_ebi)
 
-        self.K = self.n_classes = self.n_conditions_directories[self.level_classes]
-        self.n_samples = cc_maths.prod(self.n_conditions_directories)
-        # self.n_samples = math.prod(self.n_conditions_directories)
+                        self.indexes_batch[0] = i
+
+                        image_ebi = PIL.Image.open(directory_ebi)
+                        self.tensor_batch[tuple(self.indexes_batch)] = self.transforms(image_ebi)
+
+                    yield [self.tensor_batch, self.labels_eb, self.combinations_eb]
+
+        elif self.return_directories_eb:
+
+            for b in range(self.n_batches):
+                self.labels_eb = self.labels[self.batches_indexes[b]]
+                # batch_size_eb = batches_indexes_e.shape[0]
+                for i in range(self.batch_size):
+                    combination_directory_ebi = self.combinations_directories[self.batches_indexes[b][i], :]
+                    combination_directory_str_ebi = [
+                        self.conditions_directories_names[l][combination_directory_ebi[l]]
+                        for l in range(self.L)]
+                    directory_ebi = os.path.join(self.directory_root, *combination_directory_str_ebi)
+
+                    self.directories_eb[i] = directory_ebi
+
+                    self.indexes_batch[0] = i
+
+                    image_ebi = PIL.Image.open(directory_ebi)
+                    self.tensor_batch[tuple(self.indexes_batch)] = self.transforms(image_ebi)
+
+                yield [self.tensor_batch, self.labels_eb, self.directories_eb]
 
-        if batch_size is None:
-            if n_batches is None:
-                self.batch_size = self.n_samples
-                self.n_batches = 1
-            else:
-                self.n_batches = n_batches
-                self.batch_size = cc_maths.rint(self.n_samples / self.n_batches)
         else:
-            self.batch_size = batch_size
-            self.n_batches = cc_maths.rint(self.n_samples / self.batch_size)
 
-        print('self.batch_size =', self.batch_size)
-        print('self.n_samples =', self.n_samples)
-        print('self.n_batches =', self.n_batches)
-        print('self.time.range =', self.time.range)
-        print('self.n_conditions_directories =', self.n_conditions_directories)
-        # print('self.conditions_directories =', self.conditions_directories)
+            for b in range(self.n_batches):
+                self.labels_eb = self.labels[self.batches_indexes[b]]
+                # batch_size_eb = batches_indexes_e.shape[0]
+                for i in range(self.batch_size):
+                    combination_directory_ebi = self.combinations_directories[self.batches_indexes[b][i], :]
+                    combination_directory_str_ebi = [
+                        self.conditions_directories_names[l][combination_directory_ebi[l]]
+                        for l in range(self.L)]
+                    directory_ebi = os.path.join(self.directory_root, *combination_directory_str_ebi)
+                    self.indexes_batch[0] = i
 
-        if self.shuffle:
-            self.batches_indexes = None
-        else:
-            self.batches_indexes = np.split(np.arange(self.n_samples), self.n_batches, axis=0)
+                    image_ebi = PIL.Image.open(directory_ebi)
+                    self.tensor_batch[tuple(self.indexes_batch)] = self.transforms(image_ebi)
 
-        if self.time.variate_shift:
-            self.combinations_directories_no_shift = (
-                cc_combinations.conditions_to_combinations(self.conditions_directories))
-            self.combinations_directories = np.copy(self.combinations_directories_no_shift)
-            self.combinations_directories[:, self.time.level] += self.time.shift
-        else:
-            self.combinations_directories_no_shift = None
-            self.combinations_directories = (
-                cc_combinations.conditions_to_combinations(self.conditions_directories))
-
-        self.labels = torch.tensor(
-            self.combinations_directories[slice(0, self.n_samples, 1), self.level_classes],
-            dtype=torch.int64, device=device)
-
-        combination_directory_str_0 = [self.conditions_directories_names[v][0] for v in range(self.L)]
-        directory_0 = os.path.join(self.directory_root, *combination_directory_str_0)
-        image_0 = PIL.Image.open(directory_0)
-        tensor_0 = self.transforms(image_0)
-        # shape_sample_0 = np.asarray(tensor_0.shape)
-        # self.shape_sample = shape_sample_0
-        # self.shape_batch = np.empty(self.shape_sample.size + 1, self.shape_sample.dtype)
-        # self.shape_batch[0] = self.batch_size
-        # self.shape_batch[1:] = self.shape_sample
-        # self.tensor_batch = torch.empty(tuple(self.shape_batch), dtype=tensor_0.dtype)
-        shape_sample_0 = list(tensor_0.shape)
-        self.shape_sample = shape_sample_0
-        self.shape_batch = torch.Size(tuple([self.batch_size] + shape_sample_0))
-        self.tensor_batch = torch.empty(self.shape_batch, dtype=tensor_0.dtype, device=device)
-
-        self.n_dims_samples = self.shape_sample.__len__()
-        self.n_dims_batch = self.n_dims_samples + 1
-        self.indexes_batch = np.empty(self.n_dims_batch, dtype=object)
-        self.indexes_batch[1:] = slice(0, None, 1)
-
-        print('self.tensor_batch.shape =', self.tensor_batch.shape)
-        print()
-
-    def load_batches_e(self):
-
-        if self.shuffle:
-            self.batches_indexes = np.split(np.random.permutation(self.n_samples), self.n_batches, axis=0)
-
-        # for b in range(3):
-        for b in range(self.n_batches):
-            labels_b = self.labels[self.batches_indexes[b]]
-            # batch_size_eb = batches_indexes_e.shape[0]
-            for i in range(self.batch_size):
-                combination_directory_ebi = self.combinations_directories[self.batches_indexes[b][i], :]
-                combination_directory_str_ebi = [
-                    self.conditions_directories_names[l][combination_directory_ebi[l]]
-                    for l in range(self.L)]
-                directory_ebi = os.path.join(self.directory_root, *combination_directory_str_ebi)
-                self.indexes_batch[0] = i
-
-                image_ebi = PIL.Image.open(directory_ebi)
-                self.tensor_batch[tuple(self.indexes_batch)] = self.transforms(image_ebi)
-
-            yield [self.tensor_batch, labels_b]
-
-        if (self.time is not None) and self.time.variate_shift:
-            self.time.next_shift()
-            self.combinations_directories = np.copy(self.combinations_directories_no_shift)
-            self.combinations_directories[:, self.time.level] += self.time.shift
+                yield [self.tensor_batch, self.labels_eb]
 
 
 def train_early_stop(
@@ -617,7 +520,7 @@ def train_early_stop(
                 os.remove(directory_stats)
             cc_txt.lines_to_csv_file(directory_stats, stats['lines'], stats['headers'])
 
-        print('Epoch: {:d}. Validation. Loss: {:s}. Lowest Loss {:s}. Accuracy: {:s}. Highest Accuracy: {:s}.'.format(
+        print('Epoch: {:d}. Validation. Loss: {:s}. Lowest Loss: {:s}. Accuracy: {:s}. Highest Accuracy: {:s}.'.format(
             e, loss_str_e, lowest_loss_str, accuracy_str_e, highest_accuracy_str))
 
         print('Epoch {e} - Unsuccessful Epochs {i}.'.format(e=e, i=i))
@@ -630,7 +533,7 @@ def train_early_stop(
 
     time_training = cc_timer.get_delta_time()
 
-    print('Training complete in {d} days {h} hours {m} minutes {s} seconds'.format(
+    print('Training completed in {d} days {h} hours {m} minutes {s} seconds'.format(
         d=time_training.days, h=time_training.hours,
         m=time_training.minutes, s=time_training.seconds))
     print('Number of Epochs: {E:d}'.format(E=E))
@@ -642,9 +545,111 @@ def train_early_stop(
     return model, stats
 
 
-def load_model(name_model, K=None, pretrained=False, device=None):
+def test_classifier(model, criterion, loader):
+    loader = LoaderOfBalancedAndClassDirectoryDataset()
+    # timer = cc_clock.Timer()
+    cc_timer = cc_clock.Timer()
+
+    headers_stats = ['N_samples', *['C_' + str(v) for v in range(loader.V)], 'Loss', 'Accuracy']
+
+    n_columns_stats = len(headers_stats)
+    line_stats = [loader.n_samples, *loader.n_conditions_directories, None, None]  # type: list
+
+    stats = {
+        'headers': {headers_stats[i]: i for i in range(n_columns_stats)},
+        'line': line_stats}
+
+    headers_trials = [
+        'Trial',
+        'Label',
+        *['Condition_' + str(v) for v in range(loader.V)],
+        *['Probability_' + str(k) for k in range(loader.K)],
+        'Classification',
+        'Correct_Classification'
+    ]
+
+    n_columns_trials = len(headers_trials)
+
+    trials = {
+        'headers': {headers_trials[i]: i for i in range(n_columns_trials)},
+        'lines': np.empty([loader.n_samples, n_columns_trials], dtype=object)}
+
+    n_decimals_for_printing = 6
+
+    if model2.training:
+        model.eval()  # Set model to evaluate mode
+
+    torch.set_grad_enabled(False)
+
+    running_loss_e = 0.0
+    running_corrects_e = 0
+
+    b = 0
+    # Iterate over data.
+    for data_eb in loader.load_batches_e():
+
+        samples_eb, labels_eb, combinations_eb, directories_eb = data_eb
+
+        # forward
+        outputs = model(samples_eb)
+        _, preds = torch.max(outputs, 1)
+        loss_eb = criterion(outputs, labels_eb)
+        # todo: get probabilities
+        # todo: fill trials['lines']
+
+        running_loss_e += loss_eb.item() * samples_eb.shape[0]
+        # noinspection PyTypeChecker
+        running_corrects_e += torch.sum(preds == labels_eb).item()
+
+        b += 1
+
+    loss_e = running_loss_e / loader['validation'].n_samples
+    accuracy_e = running_corrects_e / loader['validation'].n_samples
+    # loss_e = float(running_loss_e / loader['training'].n_samples)
+    # accuracy_e = float(running_corrects_e / loader['training'].n_samples)
+
+    loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
+    accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)
+
+    stats['line'][stats['headers']['Loss']] = loss_e
+    stats['line'][stats['headers']['Accuracy']] = accuracy_e
+
+    print('Test. Loss: {:s}. Accuracy: {:s}.'.format(loss_str_e, accuracy_str_e))
+
+    print()
+
+    time_training = cc_timer.get_delta_time()
+
+    print('Test completed in {d} days {h} hours {m} minutes {s} seconds'.format(
+        d=time_training.days, h=time_training.hours,
+        m=time_training.minutes, s=time_training.seconds))
+
+    return stats, trials
+
+
+def load_resnet(name_model, K=None, softmax=False, pretrained=False, device=None):
+
+    model = load_model(name_model, pretrained=pretrained, device=None)
+    if (K is None) or (K == model.fc.out_features):
+        if softmax:
+            model.fc = torch.nn.Sequential(model.fc, torch.nn.Softmax())
+    else:
+        num_ftrs = model.fc.in_features
+        if softmax:
+            model.fc = torch.nn.Sequential(torch.nn.Linear(num_ftrs, K), torch.nn.Softmax())
+        else:
+            # Here the size of each output sample is set to K.
+            model.fc = torch.nn.Linear(num_ftrs, K)
+
+    model = set_device(model, device)
+
+    return model
+
+
+def load_model(name_model, pretrained=False, device=None):
     # model = torch.hub.load('pytorch/vision:v0.6.0', 'resnet152', pretrained=True)
 
+    # model = torchvision.models.resnet18()
     if isinstance(name_model, str):
         template_load_string = f"model = torchvision.models.{name_model:s}(pretrained=pretrained)"
     else:
@@ -653,19 +658,23 @@ def load_model(name_model, K=None, pretrained=False, device=None):
     dict_globals = {'__builtins__': None}
     dict_locals = {'torchvision': torchvision, 'pretrained': pretrained}  # type: dict
     exec(template_load_string, dict_globals, dict_locals)
-    model = dict_locals['model']  # type: torchvision.models.ResNet
+    model = dict_locals['model']
 
-    if K is not None:
-        num_ftrs = model.fc.in_features
-        # Here the size of each output sample is set to K.
-        model.fc = torch.nn.Linear(num_ftrs, K)
-
-    if device is not None:
-        if isinstance(device, str):
-            if device != 'cpu':
-                model = model.to(torch.device(device))
-        elif isinstance(device, torch.device):
-            if device.type != 'cpu':
-                model = model.to(device)
+    model = set_device(model, device)
 
     return model
+
+
+def set_device(tensor, device):
+
+    if device is None:
+        return tensor
+    elif isinstance(device, str):
+        if device != 'cpu':
+            tensor = tensor.to(torch.device(device))
+        return tensor
+    elif isinstance(device, torch.device):
+        if device.type != 'cpu':
+            tensor = tensor.to(device)
+        return tensor
+
